import logging
import threading
import uuid
from datetime import datetime

import jsonpickle as jsonpickle
from elasticsearch import Elasticsearch


class Elastic:
    def __init__(self, component: str, host: str = 'localhost', port: int = 9200):
        self.es = Elasticsearch(hosts=host, port=port)
        self.component = component

    def post(self, severity, message, args):
        try:
            body = {"message": message, "@timestamp": datetime.utcnow(), "application": self.component}

            if severity:
                body["severity"] = str(severity)

            if args:
                body["args"] = args

            self.es.index(index="logs-" + '{0:%Y-%m-%d}'.format(datetime.now()), doc_type="_doc", document=body)
            pass
        except Exception as e:
            print("Elasticsearch unavailable: An exception occurred [", e, "]")
            pass

    def async_post(self, severity, message, args):
        threading.Thread(target=self.post, args=(severity, message, args)).start()


class Logger:

    def __init__(self, component, _id=None, host: str = 'localhost', port: int = 9200):
        if not _id:
            self.generate_id()
        else:
            self.id = str(_id)
        self.component = component
        log = logging.getLogger('internal')
        log.setLevel(logging.DEBUG)
        self.elastic = Elastic(self.component, host, port)
        self.log = log

    def generate_id(self):
        self.id = str(uuid.uuid1().time)

    def info(self, message, *args):
        self.logger(logging.INFO, message, args)

    def debug(self, message, *args):
        self.logger(logging.DEBUG, message, args)

    def warning(self, message, *args):
        self.logger(logging.WARNING, message, args)

    def critical(self, message, *args):
        self.logger(logging.CRITICAL, message, args)

    def error(self, message, *args):
        self.logger(logging.ERROR, message, args)

    def logger(self, level, message, *args):
        args_message = ''

        if args:
            args = jsonpickle.encode(args)
            args_message = " - args:[ " + args + " ]"

        self.log.log(level, self.id + ' - ' + message + args_message)
        self.elastic.async_post(logging.getLevelName(level), message, args)


if __name__ == "__main__":
    logger = Logger('test')
    e = Exception("Test Exception")
    logger.debug("This is a debug")
    logger.info("This is an info")
    logger.error("This is an error", e)



from setuptools import setup

setup(
    name='cbpcommon',
    version='0.0.15',
    url='https://github.com/Crypto-Bot-Platform/cbpcommon',
    license='MIT',
    author='Boris Tsekinovsky',
    author_email='t.boris@gmail.com',
    description='Common library for Crypto Bot Platform',
    packages=['.'],
    include_package_data=True,
    install_requires=["avro==1.11.0", "confluent_kafka==1.8.2", "setuptools>=60.9.0", "elasticsearch~=7.16.3",
                      "jsonpickle"],
)



import datetime
import io
import threading
import time
from os import environ

import avro.io
import avro.schema
from confluent_kafka import Producer, Consumer, KafkaError
from confluent_kafka.admin import AdminClient, NewTopic, NewPartitions

from eslogger import Logger


class MessageException(Exception):
    def __init__(self, m):
        super(m)


class EventManager:
    def __init__(self, host: str = 'localhost', port: int = 9092):
        self.kafka_address = host + ':' + str(port)
        self.adminClient = AdminClient({
            "bootstrap.servers": self.kafka_address
        })
        elastic_host = environ['ELASTIC_HOST'] if "ELASTIC_HOST" in environ else 'localhost'
        elastic_port = environ['ELASTIC_PORT'] if "ELASTIC_PORT" in environ else '9200'
        self.log = Logger('event-manager', host=elastic_host, port=int(elastic_port))

    def create_address(self, address: str):
        if address not in self.adminClient.list_topics().topics:
            fs = self.adminClient.create_topics([NewTopic(address, 1, 1)])
            for topic, f in fs.items():
                try:
                    f.result()  # The result itself is None
                    self.log.info("Topic {} created".format(topic))
                except Exception as e:
                    self.log.error("Failed to create topic {}: {}".format(topic, e))

    def modify_mailbox_size(self, address: str, size: int):
        if address in self.adminClient.list_topics().topics:
            fs = self.adminClient.create_partitions([NewPartitions(address, size)])
            for topic, f in fs.items():
                try:
                    f.result()  # The result itself is None
                    self.log.info("Additional partitions created for topic {}".format(topic))
                except Exception as e:
                    self.log.error("Failed to add partitions to topic {}: {}".format(topic, e))
            return fs

    def delete_address(self, address: str):
        if address not in self.adminClient.list_topics().topics:
            self.log.warning(f"Address {address} was not found")
            return
        fs = self.adminClient.delete_topics([address])
        # Wait for operation to finish.
        for topic, f in fs.items():
            try:
                f.result()  # The result itself is None
                self.log.info("Topic {} deleted".format(topic))
            except Exception as e:
                self.log.error("Failed to delete topic {}: {}".format(topic, e))

    def send_command_to_address(self, address: str, schema_str: str, command_object: object):
        record_schema = avro.schema.parse(schema_str)
        conf = {'bootstrap.servers': self.kafka_address}

        producer = Producer(**conf)

        try:
            writer = avro.io.DatumWriter(record_schema)
            bytes_writer = io.BytesIO()
            encoder = avro.io.BinaryEncoder(bytes_writer)
            writer.write(command_object, encoder)
            raw_bytes = bytes_writer.getvalue()
            producer.produce(address, raw_bytes)
            producer.flush()
        except ValueError:
            print(f"Error Invalid input {command_object}, discarding record...")

    def wait_for_command(self, address: str, schema_str: str, on):

        conf = {'bootstrap.servers': self.kafka_address,
                'group.id': 'command.query',
                'default.topic.config': {'auto.offset.reset': 'smallest'}}

        consumer = Consumer(**conf)
        consumer.subscribe([address])

        schema = avro.schema.parse(schema_str)

        try:
            while True:
                msg = consumer.poll()
                if msg is None:
                    continue
                if msg.error():
                    if msg.error().code() == KafkaError.PARTITION_EOF:
                        self.log.error('ERROR: %% %s [%d] reached end at offset %d\n' %
                                       (msg.topic(), msg.partition(),
                                        msg.offset()))
                    elif msg.error():
                        self.log.error(f"ERROR: {msg.error()}")
                else:
                    message = msg.value()
                    bytes_reader = io.BytesIO(message)
                    decoder = avro.io.BinaryDecoder(bytes_reader)
                    reader = avro.io.DatumReader(schema)
                    try:
                        decoded_msg = reader.read(decoder)
                        # on(decoded_msg)
                        threading.Thread(target=on, args=(decoded_msg,)).start()
                    except AssertionError as e:
                        self.log.error(f"ERROR: {e}")

        except Exception as e:
            self.log.error(f"Stopped polling for address {address}. Error: {e}")


if __name__ == "__main__":
    em = EventManager()
    em.create_address('unit-test')
    em.modify_mailbox_size('unit-test', 3)
    schema = """
    {
        "namespace": "confluent.io.examples.serialization.avro",
        "name": "Test",
        "type": "record",
        "fields": [
            {"name": "timestamp", "type": "long", "logicalType": "timestamp-millis"},
            {"name": "name", "type": "string"},
            {"name": "value", "type": "string"}            
        ]
    }
    """


    def send_commands():
        count = 0
        while count < 10:
            em.send_command_to_address('unit-test', schema, {
                "timestamp": int(datetime.datetime.timestamp(datetime.datetime.now())),
                "name": "Test",
                "value": str(datetime.datetime.now())})
            time.sleep(10)
            count += 1


    x = threading.Thread(target=send_commands)
    x.start()


    def handler(obj):
        print(datetime.datetime.timestamp(datetime.datetime.now()))
        print(obj['timestamp'])
        print(f"{datetime.datetime.now()}: {obj}")


    y = threading.Thread(target=em.wait_for_command, args=('unit-test', schema, handler,))
    y.start()

    time.sleep(100)

    em.delete_address("unit-test")
